local THIS_ADDON = "WatchIt"

local BOX_W = 230
local BOX_H = 32
local MAX_CHILDREN = 7

do
    local function GetTableValue (tblName, keys)
        local tmp = _G[tblName]
        if type(tmp)~='table' then
            return nil
        end
        
        for i,key in ipairs(keys) do
            tmp = tmp[key]
        end
        
        return tmp
    end

    local function CreateTextbox (parent)
        local f = parent:CreateFontString(nil, "HIGH", "GameFontNormal")
        
        --local f = CreateFrame ("EditBox", nil, parent)
        f:SetTextColor (1.0,0.77,0.27,1.0)
        --f:SetFont ('Fonts\\ARIALN.TTF', 14, "")
        --f:SetTextInsets (8, 8, 2, 2)
        f:SetFont ("Fonts\\FRIZQT__.TTF", 10)
        --f:SetAutoFocus (false)
        --f:Disable ()
        
        return f
    end

    function CreateWatchBox (idx)
        local f = CreateFrame ("Frame", 'frmWatch'..tostring(idx), UIParent)
        f.idx = idx
        f:SetSize (BOX_W, BOX_H)
        f:SetBackdrop ({
            bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
            edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
            edgeSize = 14,
        })

        -- Textboxes
        f.txtExpand = CreateTextbox(f)
        f.txtExpand:SetPoint ("TOPRIGHT", f, "TOPRIGHT")
        f.txtExpand:SetPoint ("BOTTOMRIGHT", f, "BOTTOMRIGHT")
        f.txtExpand:SetFont ("Fonts\\FRIZQT__.TTF", 16)
        f.txtExpand:SetText ('+')
        f.txtExpand:SetJustifyH ('CENTER')
        
        f.txtLeft = CreateTextbox(f)
        f.txtLeft:SetPoint ("TOPLEFT", f, "TOPLEFT", 4, 0)
        f.txtLeft:SetPoint ("BOTTOMLEFT", f, "BOTTOMLEFT", 4, 0)
        f.txtLeft:SetWidth (76)        
        
        f.txtRight = CreateTextbox(f)
        f.txtRight:SetPoint ("TOPLEFT", f.txtLeft, "TOPRIGHT")
        f.txtRight:SetPoint ("BOTTOMRIGHT", f.txtExpand, "BOTTOMLEFT")
        f.txtRight:SetJustifyH ("LEFT")
        
        -- Create up/down buttons, keep them hidden till they're needed
        f.btnUp = CreateFrame ("Button", nil, f)
        f.btnUp:SetSize (BOX_W - 10, 16)
        f.btnUp:SetNormalTexture ("Interface\\Addons\\"..THIS_ADDON.."\\up")
        f.btnUp:SetHighlightTexture ("Interface\\Buttons\\UI-Common-MouseHilight", "ADD")
        f.btnUp:Hide()
        f.btnUp:SetScript ("OnClick", function (self)
            self:GetParent():ScrollUp()
        end)
        f.btnDown = CreateFrame ("Button", nil, f)
        f.btnDown:SetSize (BOX_W - 10, 16)
        f.btnDown:SetNormalTexture ("Interface\\Addons\\"..THIS_ADDON.."\\down")
        f.btnDown:SetHighlightTexture ("Interface\\Buttons\\UI-Common-MouseHilight", "ADD")
        f.btnDown:Hide()
        f.btnDown:SetScript ("OnClick", function (self)
            self:GetParent():ScrollDown()
        end)
        
        ------------------------------
        
        function f:GetRootFrame ()
            if self:GetParent()==UIParent then
                return self
            else
                return self:GetParent():GetRootFrame()
            end
        end        
        ------------------------------
        
        function f:CreateTag (isMovable)
            local f = CreateFrame ("Frame", nil, self)
            f:SetSize (BOX_W-10, 22)
            f:SetPoint ("BOTTOM", self, "TOP", 0, -4)
            f:SetBackdrop ({
                bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
                edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
                edgeSize = 10,
            })
            
            -- X button
            f.btnX = CreateFrame ("EditBox", nil, f)
            f.btnX:SetPoint ("TOPRIGHT", f, "TOPRIGHT")
            f.btnX:SetPoint ("BOTTOMLEFT", f, "BOTTOMRIGHT", -13, 0)
            f.btnX:SetTextColor (1.0,0.77,0.27,1.0)
            f.btnX:SetTextInsets (0, 0, 2, 2)
            f.btnX:SetFont ("Fonts\\FRIZQT__.TTF", 10)
            f.btnX:SetAutoFocus (false)
            f.btnX:Disable ()
            f.btnX:SetText ('x')
            f.btnX:SetScript ("OnMouseDown", function (self, button)
                self:GetParent():GetParent():Free()
            end)            
            
            -- Label
            f.lbl = f:CreateFontString(nil, "HIGH", "GameFontNormal")
            f.lbl:SetPoint ("TOPLEFT", f, "TOPLEFT", 2, 0)
            f.lbl:SetPoint ("BOTTOMRIGHT", f.btnX, "BOTTOMLEFT")
            f.lbl:SetTextColor (1.0,0.77,0.27,1.0)
            f.lbl:SetFont ("Fonts\\FRIZQT__.TTF", 10)
            f.lbl:SetText (self.varName)
            
            -- Make movable... maybe
            if isMovable then
                self:SetMovable (true)
                --f:RegisterForDrag ("LeftButton")
                self:SetClampedToScreen (true)
                self:EnableMouse (true)
                self:EnableMouseWheel (true)
                f:SetScript("OnMouseDown", function(self, button)
                    if button=='LeftButton' then
                        self:GetParent():StartMoving()
                    end
                end)
                f:SetScript("OnMouseUp", function(self, button)
                    if button=='LeftButton' then
                        self:GetParent():StopMovingOrSizing()
                    end
                end)
            end
            
            self.tag = f
            return f
        end
        
        ------------------------------
        
        function f:SetTableView (isTable)
            self.isTable = isTable
            
            if isTable then
                -- List of child values
                self.childKeys = {}
                for k,v in pairs(self.value) do
                    table.insert (self.childKeys, k)
                end
                
                self.txtExpand:SetWidth(30)
                self.isExpanded = false
                self:Collapse()
            else
                self.txtExpand:SetWidth(1)
            end
        end
        ------------------------------
        
        function f:LoadVar (varName)
            self.varName = varName
            local v = _G[varName]
            self.value = v
            self.txtRight:SetText (tostring(v))
            self.txtLeft:SetText ('('..type(v)..')')
            
            self:CreateTag (true)
            self:SetTableView (type(v)=='table')
        end
        ------------------------------
        
        function f:LoadTableElement (varName, key)
            self.varName = varName
            self.keys = {}
            for i,v in ipairs(self:GetParent().keys) do
                self.keys[i] = self:GetParent().keys[i]
            end
            table.insert (self.keys, key)
            
            self.txtLeft:SetText (key)
            local v = GetTableValue (varName, self.keys)
            self.value = v
            self.txtRight:SetText (tostring(v))
            
            self:SetTableView (type(v)=='table')
        end
        ------------------------------
        
        function f:ShowChildrenFrom (first)
            self:CollapseChildren()
            
            -- Bugger counting... this loop can just go till it runs out of frames
            local cidx = 1  -- child index
            for i=first,first+#self.childKeys-1 do
                if (not self.childFrames[cidx]) or (i>#self.childKeys) then break end
                self.childFrames[cidx]:LoadTableElement (self.varName, self.childKeys[i])
                cidx = cidx + 1
            end
            
            self.firstShownChild = first

            self.btnUp:SetShown (first>1)
            self.btnDown:SetShown (first+MAX_CHILDREN-1 < #self.childKeys)
                
        end        
        ------------------------------

        function f:ScrollUp ()
            if self.btnUp:IsShown() and #self.childKeys>MAX_CHILDREN then
                self.firstShownChild = self.firstShownChild - 1
                f:ShowChildrenFrom (self.firstShownChild)
            end
        end
        
        function f:ScrollDown ()
            if self.btnDown:IsShown() and #self.childKeys>MAX_CHILDREN then
                self.firstShownChild = self.firstShownChild + 1
                f:ShowChildrenFrom (self.firstShownChild)
            end
        end
        ------------------------------
        
        function f:Expand ()
            -- Collapse all siblings
            if #self.keys > 0 then
                local p = self:GetParent()
                p:CollapseChildren()
                --p:Expand()
            end
            
            -- Wait, does this frame even have children?
            if #self.childKeys == 0 then
                GameTooltip:SetOwner (self, "ANCHOR_CURSOR")
                GameTooltip:AddLine ('This table is empty!', 1.0, 1.0, 0.8)
                GameTooltip:Show ()
                return
            end            
            
            -- Show children
            for i=1, min(MAX_CHILDREN, #self.childKeys) do
                local tmp = poolHandler.get()
                tmp:SetParent (self)
                tmp:SetPoint ("TOPLEFT", self, "TOPRIGHT", 0, -(i-1) * BOX_H)
                
                tmp:Show()
                table.insert (self.childFrames, tmp)                                
            end
            
            self:ShowChildrenFrom (self.firstShownChild or 1)
            
            -- Place up/down buttons
            if #self.childKeys > MAX_CHILDREN then
                self.btnUp:SetPoint ("BOTTOM", self.childFrames[1], "TOP")
                self.btnDown:SetPoint ("TOP", self.childFrames[MAX_CHILDREN], "BOTTOM")
            end
            
            self.txtExpand:SetText ('-')
            self.isExpanded = true
        end
        ------------------------------
        
        function f:Collapse ()
            if not self.isTable then return end
        
            self.txtExpand:SetText ('+')
            self.isExpanded = false
            
            if #self.childKeys>MAX_CHILDREN then
                self.btnUp:Hide()
                self.btnDown:Hide()
            end
            
            for i,v in ipairs(self.childFrames) do
                --v:Collapse()
                v:Free()
                self.childFrames[i] = nil
            end
        end
        ------------------------------
        
        function f:CollapseChildren ()
            for i,v in ipairs(self.childFrames) do
                self.childFrames[i]:Collapse()
            end
        end
        ------------------------------
        -- Events
        
        f:SetScript ("OnMouseDown", function (self, button,...)
            if self.isTable then
                if self.isExpanded then
                    self:Collapse ()
                else
                    self:Expand ()
                end
            end
        end)
        ------------------------------
        
        f:SetScript("OnMouseWheel", function (self, delta)
            if IsControlKeyDown() then
                self:GetRootFrame():SetScale (self:GetRootFrame():GetScale() + delta*0.04)
            else
                if delta>0 then 
                    self:GetParent():ScrollUp()
                else
                    self:GetParent():ScrollDown()
                end
            end
        end)        
        ------------------------------
        
        f:SetScript ("OnLeave", function (self)
            GameTooltip:Hide()
        end)
        ------------------------------
        
        function f:Free ()
            -- Free children
            --[[if self.isTable then
                for k,v in pairs(self.childFrames) do
                    v:Free ()
                end
            end]]
            self:Collapse()
            
            -- Get rid of the extra controls
            if self.tag then
                self.tag:Hide()
                self.tag = nil
            end
            self.btnUp:Hide()
            self.btnDown:Hide()
            
            self.varName = nil
            self.value = nil            
            
            self:SetParent (UIParent)
            self:Hide()
            
            poolHandler.free (self.idx)
        end
        
        ------------------------------
        
        function f:Init ()
            self.varName = nil
            self.keys = {}
            -- A box showing g_tbl['druid']['balance'] will have f.keys=={'druid','balance'} and f.varName='g_tbl'
            self.free = false
            self.childFrames = {}
            self.childKeys = {}
            self.isExpanded = nil
            self.firstShownChild = 1
        end
        ------------------------------
        
        function f:InitWatchFrame ()
            self:SetTableView (false)        
        end
        
        ------------------------------
        
        function f:LoadValue (v)
            self.value = v
            self.txtRight:SetText (tostring(v))
            self.txtLeft:SetText ('('..type(v)..')')
        end
        
        ------------------------------
        
        function f:AdjustWidth (w)
            local delta = self:GetWidth() - w
            self:SetWidth (w)
            self.tag:SetWidth (self.tag:GetWidth() - delta)
        end
        
        ------------------------------
        
        f:Hide()
        f:Init()
        f.free = true
        
        return f
    end
end

------------------------------------------------------------------

do
    local pool = {}
    local countFree = 0
    
    local function GetSpecificBox (n)
        if not pool[n].free then
            print ('Error: box '..tostring(n)..' is not free')
            return nil
        end
        
        pool[n]:Init()
        pool[n]:Show()
        countFree = countFree - 1
        return pool[n]
    end
    
    local function GetBox ()
        if countFree > 0 then
            for i=1,#pool do
                if pool[i].free then
                    return GetSpecificBox(i)
                end
            end
        else
            local tmp = CreateWatchBox(#pool+1)
            table.insert (pool, tmp)
            countFree = countFree + 1
            return GetSpecificBox(#pool)
        end
        
        print ('WatchIt Weird Error Code 10001')
    end
    
    local function FreeBox (idx)
        if pool[idx].free then
            --local errcode = 10000+idx
            --print ('WatchIt Weird Error Code '..errcode)
            return 
        end
        
        countFree = countFree + 1
        pool[idx].free = true
        pool[idx]:Hide()
    end
    
    local function ResetPool (idx)
        for i=1,#pool do
            pool[i]:Hide()
            pool[i] = nil
        end
        countFree = 0
    end
    
    poolHandler = {
        get = GetBox,
        free = FreeBox,
        reset = ResetPool,
        report = function ()
            print ('Pool size: '..#pool)
            print ('Free boxes: '..countFree)
        end
    }

end

------------------------------------------------------------------

function WatchVar (varName)
    local f = poolHandler.get()
    f:SetPoint ("TOPLEFT", UIParent, "TOPLEFT", 100, -150)
    f:LoadVar (varName)
    return f
end

------------------------------------------------------------------

function CreateWatchFrame (title)
    local f = poolHandler.get()
    f:SetPoint ("TOPLEFT", UIParent, "TOPLEFT", 100, -150)
    f.varName = title
    f:CreateTag (true)
    f:InitWatchFrame ()
    
    return f
end

------------------------------------------------------------------
--[[
do
    local f = CreateFrame ("Frame", 'frmWatchIt', UIParent)
    f:SetSize (BOX_W-10, 22)
    f:SetPoint ("TOP", UIParent, "TOP", 0, -15)
    f:SetBackdrop ({
        bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
        edgeSize = 10,
    })
    
    -- X button
    f.btnX = CreateFrame ("EditBox", nil, f)
    f.btnX:SetPoint ("TOPRIGHT", f, "TOPRIGHT")
    f.btnX:SetPoint ("BOTTOMLEFT", f, "BOTTOMRIGHT", -13, 0)
    f.btnX:SetTextColor (1.0,0.77,0.27,1.0)
    f.btnX:SetTextInsets (0, 0, 2, 2)
    f.btnX:SetFont ("Fonts\\FRIZQT__.TTF", 10)
    f.btnX:SetAutoFocus (false)
    f.btnX:Disable ()
    f.btnX:SetText ('x')
    f.btnX:SetScript ("OnMouseDown", function (self, button)
        frmWatchIt:Hide()
    end)            
    
    -- Label
    f.lbl = f:CreateFontString(nil, "HIGH", "GameFontNormal")
    f.lbl:SetPoint ("TOPLEFT", f, "TOPLEFT", 2, 0)
    f.lbl:SetPoint ("BOTTOMRIGHT", f.btnX, "BOTTOMLEFT")
    f.lbl:SetTextColor (1.0,0.77,0.27,1.0)
    f.lbl:SetFont ("Fonts\\FRIZQT__.TTF", 10)
    f.lbl:SetText ('WatchIt')
    
    -- Make movable
    f:SetMovable (true)
    --f:RegisterForDrag ("LeftButton")
    f:SetClampedToScreen (true)
    f:EnableMouse (true)
    f:EnableMouseWheel (true)
    f:SetScript("OnMouseDown", function(self, button)
        if button=='LeftButton' then
            self:StartMoving()
        end
    end)
    f:SetScript("OnMouseUp", function(self, button)
        if button=='LeftButton' then
            self:StopMovingOrSizing()
        end
    end)
    
    f.boxes = {[0]=f}
    function f:PlaceBoxes ()
        for i=1,#self.boxes do
            self.boxes[i]:SetPoint ("TOP", self.boxes[i-1], "BOTTOM")
        end
    end
end
]]

